from . import dynamic_api_gen
from .version import __version__
from os import path, environ

try:
    if path.exists(environ.get("PHARUS_SPEC_PATH")):
        dynamic_api_gen.populate_api()
except TypeError:
    print("No Dynamic API path found")

try:
    from .dynamic_api import app
except ImportError:
    from .server import app

__all__ = ["__version__", "app"]



"""Library for interfaces into DataJoint pipelines."""
import datajoint as dj
from datajoint.utils import to_camel_case
from datajoint.user_tables import UserTable
from datajoint import VirtualModule
import datetime
import numpy as np
import re
from .error import InvalidRestriction, UnsupportedTableType

DAY = 24 * 60 * 60
DEFAULT_FETCH_LIMIT = 1000  # Stop gap measure to deal with super large tables


class _DJConnector:
    """Primary connector that communicates with a DataJoint database server."""

    @staticmethod
    def _attempt_login(database_address: str, username: str, password: str):
        """
        Attempts to authenticate against database with given username and address.

        :param database_address: Address of database
        :type database_address: str
        :param username: Username of user
        :type username: str
        :param password: Password of user
        :type password: str
        """
        dj.config["database.host"] = database_address
        dj.config["database.user"] = username
        dj.config["database.password"] = password

        # Attempt to connect return true if successful, false is failed
        dj.conn(reset=True)

    @staticmethod
    def _list_schemas(jwt_payload: dict) -> list:
        """
        List all schemas under the database.

        :param jwt_payload: Dictionary containing databaseAddress, username, and password
            strings
        :type jwt_payload: dict
        :return: List of schemas names in alphabetical order (excludes ``information_schema``,
            ``sys``, ``performance_schema``, ``mysql``)
        :rtype: list
        """
        _DJConnector._set_datajoint_config(jwt_payload)

        # Attempt to connect return true if successful, false is failed
        return [
            row[0]
            for row in dj.conn().query(
                """
                SELECT SCHEMA_NAME FROM information_schema.schemata
                WHERE SCHEMA_NAME NOT IN (
                    "information_schema", "sys", "performance_schema", "mysql"
                )
                ORDER BY SCHEMA_NAME
                """
            )
        ]

    @staticmethod
    def _list_tables(jwt_payload: dict, schema_name: str) -> dict:
        """
        List all tables and their type given a schema.

        :param jwt_payload: Dictionary containing databaseAddress, username, and password
            strings
        :type jwt_payload: dict
        :param schema_name: Name of schema to list all tables from
        :type schema_name: str
        :return: Contains a key for each table type where values are the respective list of
            table names
        :rtype: dict
        """
        _DJConnector._set_datajoint_config(jwt_payload)

        # Get list of tables names
        tables_name = dj.Schema(schema_name, create_schema=False).list_tables()
        # Dict to store list of table name for each type
        tables_dict_list = dict(manual=[], lookup=[], computed=[], imported=[], part=[])
        # Loop through each table name to figure out what type it is and add them to
        # tables_dict_list
        for table_name in tables_name:
            table_type = dj.diagram._get_tier(
                "`" + schema_name + "`.`" + table_name + "`"
            ).__name__
            if table_type == "Manual":
                tables_dict_list["manual"].append(dj.utils.to_camel_case(table_name))
            elif table_type == "Lookup":
                tables_dict_list["lookup"].append(dj.utils.to_camel_case(table_name))
            elif table_type == "Computed":
                tables_dict_list["computed"].append(dj.utils.to_camel_case(table_name))
            elif table_type == "Imported":
                tables_dict_list["imported"].append(dj.utils.to_camel_case(table_name))
            elif table_type == "Part":
                table_name_parts = table_name.split("__")
                tables_dict_list["part"].append(
                    to_camel_case(table_name_parts[-2])
                    + "."
                    + to_camel_case(table_name_parts[-1])
                )
            else:
                raise UnsupportedTableType(table_name + " is of unknown table type")
        return tables_dict_list

    @staticmethod
    def _fetch_records(
        query,
        restriction: list = [],
        limit: int = 1000,
        page: int = 1,
        order=None,
        fetch_blobs=False,
        fetch_args=[],
    ) -> tuple:
        """
        Get records from query.

        :param query: any datajoint object related to QueryExpression
        :type query: datajoint ``QueryExpression`` or related object
        :param restriction: Sequence of filters as ``dict`` with ``attributeName``,
            ``operation``, ``value`` keys defined, defaults to ``[]``
        :type restriction: list, optional
        :param limit: Max number of records to return, defaults to ``1000``
        :type limit: int, optional
        :param page: Page number to return, defaults to ``1``
        :type page: int, optional
        :param order: Sequence to order records, defaults to ``['KEY ASC']``. See
            :class:`~datajoint.fetch.Fetch` for more info.
        :type order: list, optional
        :return: Attribute headers, records in dict form, and the total number of records that
            can be paged
        :rtype: tuple
        """

        # Get table object from name
        attributes = query.heading.attributes
        # Fetch tuples without blobs as dict to be used to create a
        #   list of tuples for returning
        query_restricted = query & dj.AndList(
            [
                _DJConnector._filter_to_restriction(
                    f, attributes[f["attributeName"]].type
                )
                for f in restriction
            ]
        )

        order_by = (
            fetch_args.pop("order_by") if "order_by" in fetch_args else ["KEY ASC"]
        )
        order_by = order if order else order_by

        limit = fetch_args.pop("limit") if "limit" in fetch_args else limit

        if fetch_blobs and not fetch_args:
            fetch_args = [*query.heading.attributes]
        elif not fetch_args:
            fetch_args = query.heading.non_blobs
        else:
            attributes = {k: v for k, v in attributes.items() if k in fetch_args}
        non_blobs_rows = query_restricted.fetch(
            *fetch_args,
            as_dict=True,
            limit=limit,
            offset=(page - 1) * limit,
            order_by=order_by,
        )

        # Buffer list to be return
        rows = []

        # Looped through each tuple and deal with TEMPORAL types and replacing
        #   blobs with ==BLOB== for json encoding
        for non_blobs_row in non_blobs_rows:
            # Buffer object to store the attributes
            row = []
            # Loop through each attributes, append to the tuple_to_return with specific
            #   modification based on data type
            for attribute_name, attribute_info in attributes.items():
                if not attribute_info.is_blob:
                    if non_blobs_row[attribute_name] is None:
                        # If it is none then just append None
                        row.append(None)
                    elif attribute_info.type == "date":
                        # Date attribute type covert to epoch time
                        row.append(
                            (
                                non_blobs_row[attribute_name]
                                - datetime.date(1970, 1, 1)
                            ).days
                            * DAY
                        )
                    elif attribute_info.type == "time":
                        # Time attirbute, return total seconds
                        row.append(non_blobs_row[attribute_name].total_seconds())
                    elif re.match(r"^datetime.*$", attribute_info.type) or re.match(
                        r"timestamp", attribute_info.type
                    ):
                        # Datetime or timestamp, use timestamp to covert to epoch time
                        row.append(non_blobs_row[attribute_name].timestamp())
                    elif attribute_info.type[0:7] == "decimal":
                        # Covert decimal to string
                        row.append(str(non_blobs_row[attribute_name]))
                    else:
                        # Normal attribute, just return value with .item to deal with numpy
                        #   types
                        if isinstance(non_blobs_row[attribute_name], np.generic):
                            row.append(np.asscalar(non_blobs_row[attribute_name]))
                        else:
                            row.append(non_blobs_row[attribute_name])
                else:
                    # Attribute is blob type thus fill it in string instead
                    (
                        row.append(non_blobs_row[attribute_name])
                        if fetch_blobs
                        else row.append("=BLOB=")
                    )
            # Add the row list to tuples
            rows.append(row)
        return list(attributes.keys()), rows, len(query_restricted)

    @staticmethod
    def _get_attributes(query) -> dict:
        """
        Method to get primary and secondary attributes of a query.

        :param query: any datajoint object related to QueryExpression
        :type query: datajoint ``QueryExpression`` or related object
        :return: Dict with keys ``attribute_headers`` and ``attributes`` containing
            ``primary``, ``secondary`` which each contain a
            ``list`` of ``tuples`` specifying: ``attribute_name``, ``type``, ``nullable``,
            ``default``, ``autoincrement``.
        :rtype: dict
        """

        query_attributes = dict(primary=[], secondary=[])
        for attribute_name, attribute_info in query.heading.attributes.items():
            if attribute_info.in_key:
                query_attributes["primary"].append(
                    (
                        attribute_name,
                        attribute_info.type,
                        attribute_info.nullable,
                        attribute_info.default,
                        attribute_info.autoincrement,
                    )
                )
            else:
                query_attributes["secondary"].append(
                    (
                        attribute_name,
                        attribute_info.type,
                        attribute_info.nullable,
                        attribute_info.default,
                        attribute_info.autoincrement,
                    )
                )

        return dict(
            attribute_headers=["name", "type", "nullable", "default", "autoincrement"],
            attributes=query_attributes,
        )

    @staticmethod
    def _get_table_definition(
        jwt_payload: dict, schema_name: str, table_name: str
    ) -> str:
        """
        Get the table definition.

        :param jwt_payload: Dictionary containing databaseAddress, username, and password
            strings
        :type jwt_payload: dict
        :param schema_name: Name of schema
        :type schema_name: str
        :param table_name: Table name under the given schema; must be in camel case
        :type table_name: str
        :return: Definition of the table
        :rtype: str
        """
        _DJConnector._set_datajoint_config(jwt_payload)

        local_values = locals()
        local_values[schema_name] = dj.VirtualModule(schema_name, schema_name)
        return getattr(local_values[schema_name], table_name).describe()

    @staticmethod
    def _insert_tuple(
        jwt_payload: dict, schema_name: str, table_name: str, tuple_to_insert: dict
    ):
        """
        Insert record as tuple into table.

        :param jwt_payload: Dictionary containing databaseAddress, username, and password
            strings
        :type jwt_payload: dict
        :param schema_name: Name of schema
        :type schema_name: str
        :param table_name: Table name under the given schema; must be in camel case
        :type table_name: str
        :param tuple_to_insert: Record to be inserted
        :type tuple_to_insert: dict
        """
        _DJConnector._set_datajoint_config(jwt_payload)

        schema_virtual_module = dj.VirtualModule(schema_name, schema_name)
        getattr(schema_virtual_module, table_name).insert(tuple_to_insert)

    @staticmethod
    def _record_dependency(
        jwt_payload: dict, schema_name: str, table_name: str, restriction: list = []
    ) -> list:
        """
        Return summary of dependencies associated with a restricted table. Will only show
        dependencies that user has access to.

        :param jwt_payload: Dictionary containing databaseAddress, username, and password
            strings
        :type jwt_payload: dict
        :param schema_name: Name of schema
        :type schema_name: str
        :param table_name: Table name under the given schema; must be in camel case
        :type table_name: str
        :param restriction: Sequence of filters as ``dict`` with ``attributeName``,
            ``operation``, ``value`` keys defined, defaults to ``[]``
        :type restriction: list
        :return: Tables that are dependent on specific records.
        :rtype: list
        """
        _DJConnector._set_datajoint_config(jwt_payload)
        virtual_module = dj.VirtualModule(schema_name, schema_name)
        table = getattr(virtual_module, table_name)
        attributes = table.heading.attributes
        # Retrieve dependencies of related to retricted
        dependencies = [
            dict(
                schema=descendant.database,
                table=descendant.table_name,
                accessible=True,
                count=len(
                    (
                        table
                        if descendant.full_table_name == table.full_table_name
                        else descendant * table
                    )
                    & dj.AndList(
                        [
                            _DJConnector._filter_to_restriction(
                                f, attributes[f["attributeName"]].type
                            )
                            for f in restriction
                        ]
                    )
                ),
            )
            for descendant in table().descendants(as_objects=True)
        ]
        return dependencies

    @staticmethod
    def _update_tuple(
        jwt_payload: dict, schema_name: str, table_name: str, tuple_to_update: dict
    ):
        """
        Update record as tuple into table.

        :param jwt_payload: Dictionary containing databaseAddress, username, and password
            strings
        :type jwt_payload: dict
        :param schema_name: Name of schema
        :type schema_name: str
        :param table_name: Table name under the given schema; must be in camel case
        :type table_name: str
        :param tuple_to_update: Record to be updated
        :type tuple_to_update: dict
        """
        conn = _DJConnector._set_datajoint_config(jwt_payload)

        schema_virtual_module = dj.VirtualModule(schema_name, schema_name)
        with conn.transaction:
            [
                getattr(schema_virtual_module, table_name).update1(t)
                for t in tuple_to_update
            ]

    @staticmethod
    def _delete_records(
        jwt_payload: dict,
        schema_name: str,
        table_name: str,
        restriction: list = [],
        cascade: bool = False,
    ):
        """
        Delete a specific record based on the restriction given.

        :param jwt_payload: Dictionary containing databaseAddress, username, and password
            strings
        :type jwt_payload: dict
        :param schema_name: Name of schema
        :type schema_name: str
        :param table_name: Table name under the given schema; must be in camel case
        :type table_name: str
        :param restriction: Sequence of filters as ``dict`` with ``attributeName``,
            ``operation``, ``value`` keys defined, defaults to ``[]``
        :type restriction: list, optional
        :param cascade: Allow for cascading delete, defaults to ``False``
        :type cascade: bool, optional
        """
        _DJConnector._set_datajoint_config(jwt_payload)

        schema_virtual_module = dj.VirtualModule(schema_name, schema_name)

        # Get table object from name
        table = _DJConnector._get_table_object(schema_virtual_module, table_name)
        attributes = table.heading.attributes
        restrictions = [
            _DJConnector._filter_to_restriction(f, attributes[f["attributeName"]].type)
            for f in restriction
        ]

        # Compute restriction
        query = table & dj.AndList(restrictions)
        # Check if there is only 1 tuple to delete otherwise raise error
        if len(query) == 0:
            raise InvalidRestriction("Nothing to delete")

        # All check pass thus proceed to delete
        query.delete(safemode=False) if cascade else query.delete_quick()

    @staticmethod
    def _get_table_object(
        schema_virtual_module: VirtualModule, table_name: str
    ) -> UserTable:
        """
        Helper method for getting the table object based on the table name provided.

        :param schema_virtual_module: Virtual module for accesing the schema
        :type schema_virtual_module: :class:`~datajoint.schemas.VirtualModule`
        :param table_name: Name of the table; for part it should be ``Parent.Part``
        :type table_name: str
        :return: DataJoint table object.
        :rtype: :class:`~datajoint.user_tables.UserTable`
        """
        # Split the table name by '.' for dealing with part tables
        table_name_parts = table_name.split(".")
        if len(table_name_parts) == 2:
            return getattr(
                getattr(schema_virtual_module, table_name_parts[0]), table_name_parts[1]
            )
        else:
            return getattr(schema_virtual_module, table_name_parts[0])

    @staticmethod
    def _filter_to_restriction(attribute_filter: dict, attribute_type: str) -> str:
        """
        Convert attribute filter to a restriction.

        :param attribute_filter: A filter as ``dict`` with ``attributeName``, ``operation``,
            ``value`` keys defined, defaults to ``[]``
        :type attribute_filter: dict
        :param attribute_type: Attribute type
        :type attribute_type: str
        :return: DataJoint-compatible restriction
        :rtype: str
        """
        if attribute_filter["operation"] in (">", "<", ">=", "<="):
            operation = attribute_filter["operation"]
        elif attribute_filter["value"] is None:
            operation = " IS " if attribute_filter["operation"] == "=" else " IS NOT "
        else:
            operation = attribute_filter["operation"]

        if (
            isinstance(attribute_filter["value"], str)
            and not attribute_filter["value"].isnumeric()
        ):
            value = (
                f"X'{attribute_filter['value'].replace('-', '')}'"
                if attribute_type == "uuid"
                else f"'{attribute_filter['value']}'"
            )
        else:
            value = (
                "NULL"
                if attribute_filter["value"] is None
                else attribute_filter["value"]
            )
        return f"{attribute_filter['attributeName']}{operation}{value}"

    @staticmethod
    def _set_datajoint_config(jwt_payload: dict) -> dj.connection.Connection:
        """
        Method to set credentials for database.

        :param jwt_payload: Dictionary containing databaseAddress, username, and password
            strings
        :type jwt_payload: dict
        :return: DataJoint connection object.
        :rtype: :class:`~datajoint.connection.Connection`
        """
        dj.config["database.host"] = jwt_payload["databaseAddress"]
        dj.config["database.user"] = jwt_payload["username"]
        dj.config["database.password"] = jwt_payload["password"]
        return dj.conn(reset=True)



"""Error class library."""


class UnsupportedTableType(Exception):
    """Exception raised on unsupported table types."""

    pass


class InvalidRestriction(Exception):
    """Exception raised when restrictions result in no records when expected at least one."""

    pass



"""This module is a GUI component library of various common interfaces."""
import json
from base64 import b64decode
import datajoint as dj
import re
import inspect
from datetime import date, datetime
from flask import request, send_file
from .interface import _DJConnector
import os
from pathlib import Path
import types
import io
import numpy as np


class NumpyEncoder(json.JSONEncoder):
    """teach json to dump datetimes, etc"""

    npmap = {
        np.bool_: bool,
        np.uint8: int,
        np.uint16: int,
        np.uint32: int,
        np.uint64: int,
        np.int8: int,
        np.int16: int,
        np.int32: int,
        np.int64: int,
        np.float32: float,
        np.float64: float,
        np.ndarray: list,
    }

    def default(self, o):
        if type(o) in self.npmap:
            return self.npmap[type(o)](o)
        if type(o) in (datetime, date):
            return o.isoformat()
        return json.JSONEncoder.default(self, o)

    @classmethod
    def dumps(cls, obj):
        return json.dumps(obj, cls=cls)


class QueryComponent:
    attributes_route_format = None

    def __init__(self, name, component_config, static_config, jwt_payload: dict):
        lcls = locals()
        self.name = name
        if static_config:
            self.static_variables = types.MappingProxyType(static_config)
        if not all(k in component_config for k in ("x", "y", "height", "width")):
            self.mode = "dynamic"
        else:
            self.mode = "fixed"
            self.x = component_config["x"]
            self.y = component_config["y"]
            self.height = component_config["height"]
            self.width = component_config["width"]
        self.type = component_config["type"]
        self.route = component_config["route"]
        exec(component_config["dj_query"], globals(), lcls)
        self.dj_query = lcls["dj_query"]
        if self.attributes_route_format:
            self.attribute_route = self.attributes_route_format.format(
                route=component_config["route"]
            )
        if "restriction" in component_config:
            exec(component_config["restriction"], globals(), lcls)
            self.dj_restriction = lcls["restriction"]
        else:
            self.dj_restriction = lambda: dict()

        self.vm_list = [
            dj.VirtualModule(
                s,
                s,
                connection=dj.conn(
                    host=jwt_payload["databaseAddress"],
                    user=jwt_payload["username"],
                    password=jwt_payload["password"],
                    reset=True,
                ),
            )
            for s in inspect.getfullargspec(self.dj_query).args
        ]

    @property
    def fetch_metadata(self):
        return self.dj_query(*self.vm_list)

    @property
    def restriction(self):
        # first element includes the spec's restriction,
        # second element includes the restriction from query parameters
        return dj.AndList(
            [
                self.dj_restriction(),
                {
                    k: (
                        datetime.fromtimestamp(float(v)).isoformat()
                        if re.match(
                            r"^date.*$",
                            self.fetch_metadata["query"].heading.attributes[k].type,
                        )
                        else v
                    )
                    for k, v in request.args.items()
                    if k in self.fetch_metadata["query"].heading.attributes
                },
            ]
        )


class TableComponent(QueryComponent):
    attributes_route_format = "{route}/attributes"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.frontend_map = {
            "source": "sci-viz/src/Components/Table/TableView.tsx",
            "target": "TableView",
        }
        self.response_examples = {
            "dj_query_route": {
                "recordHeader": ["subject_uuid", "session_start_time", "session_uuid"],
                "records": [
                    [
                        "00778394-c956-408d-8a6c-ca3b05a611d5",
                        1565436299.0,
                        "fb9bdf18-76be-452b-ac4e-21d5de3a6f9f",
                    ],
                    [
                        "00778394-c956-408d-8a6c-ca3b05a611d5",
                        1565601663.0,
                        "d47e9a4c-18dc-4d4d-991c-d30059ec2cbd",
                    ],
                ],
                "totalCount": 9141,
            },
            "attributes_route": {
                "attributeHeaders": [
                    "name",
                    "type",
                    "nullable",
                    "default",
                    "autoincrement",
                ],
                "attributes": {
                    "primary": [
                        ["subject_uuid", "uuid", False, None, False],
                        ["session_start_time", "datetime", False, None, False],
                    ],
                    "secondary": [["session_uuid", "uuid", False, None, False]],
                },
            },
        }

    # Returns the result of djquery with paging, sorting, filtering
    def dj_query_route(self):
        fetch_metadata = self.fetch_metadata
        record_header, table_records, total_count = _DJConnector._fetch_records(
            query=fetch_metadata["query"] & self.restriction[0],
            fetch_args=fetch_metadata["fetch_args"],
            **{
                k: (
                    int(v)
                    if k in ("limit", "page")
                    else (
                        v.split(",")
                        if k == "order"
                        else json.loads(b64decode(v.encode("utf-8")).decode("utf-8"))
                    )
                )
                for k, v in request.args.items()
            },
        )
        return dict(
            recordHeader=record_header, records=table_records, totalCount=total_count
        )

    def attributes_route(self):
        attributes_meta = _DJConnector._get_attributes(self.fetch_metadata["query"])
        return dict(
            attributeHeaders=attributes_meta["attribute_headers"],
            attributes=attributes_meta["attributes"],
        )


class MetadataComponent(TableComponent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.frontend_map = "sci-viz/src/Components/Table/Metadata.tsx:Metadata"
        self.response_examples = {
            "dj_query_route": {
                "recordHeader": ["subject_uuid", "session_start_time", "session_uuid"],
                "records": [
                    [
                        "00778394-c956-408d-8a6c-ca3b05a611d5",
                        1565436299.0,
                        "fb9bdf18-76be-452b-ac4e-21d5de3a6f9f",
                    ]
                ],
                "totalCount": 1,
            },
            "attributes_route": {
                "attributeHeaders": [
                    "name",
                    "type",
                    "nullable",
                    "default",
                    "autoincrement",
                ],
                "attributes": {
                    "primary": [
                        ["subject_uuid", "uuid", False, None, False],
                        ["session_start_time", "datetime", False, None, False],
                    ],
                    "secondary": [["session_uuid", "uuid", False, None, False]],
                },
            },
        }

    def dj_query_route(self):
        fetch_metadata = self.fetch_metadata
        record_header, table_records, total_count = _DJConnector._fetch_records(
            query=fetch_metadata["query"] & self.restriction,
            fetch_args=fetch_metadata["fetch_args"],
        )
        return dict(
            recordHeader=record_header, records=table_records, totalCount=total_count
        )


class PlotPlotlyStoredjsonComponent(QueryComponent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.frontend_map = {
            "source": "sci-viz/src/Components/Plots/FullPlotly.tsx",
            "target": "FullPlotly",
        }
        self.response_examples = {
            "dj_query_route": {
                "data": [
                    {
                        "x": ["giraffes", "orangutans", "monkeys"],
                        "y": [20, 14, 23],
                        "type": "bar",
                    }
                ],
                "layout": {"title": "Total Number of Animals"},
            },
        }

    def dj_query_route(self):
        fetch_metadata = self.fetch_metadata
        return NumpyEncoder.dumps(
            (fetch_metadata["query"] & self.restriction).fetch1(
                *fetch_metadata["fetch_args"]
            )
        )


class BasicQuery(QueryComponent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.frontend_map = {
            "source": "sci-viz/src/Components/Plots/FullPlotly.tsx",
            "target": "FullPlotly",
        }
        self.response_examples = {
            "dj_query_route": {
                "recordHeader": ["subject_uuid", "session_start_time", "session_uuid"],
                "records": [
                    [
                        "00778394-c956-408d-8a6c-ca3b05a611d5",
                        1565436299.0,
                        "fb9bdf18-76be-452b-ac4e-21d5de3a6f9f",
                    ]
                ],
                "totalCount": 1,
            },
        }

    def dj_query_route(self):
        fetch_metadata = self.fetch_metadata
        record_header, table_records, total_count = _DJConnector._fetch_records(
            query=fetch_metadata["query"] & self.restriction,
            fetch_args=fetch_metadata["fetch_args"],
        )
        return dict(
            recordHeader=record_header, records=table_records, totalCount=total_count
        )


class FileImageAttachComponent(QueryComponent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.frontend_map = {
            "source": "sci-viz/src/Components/Plots/Image.tsx",
            "target": "Image",
        }
        self.response_examples = {
            "dj_query_route": b"PNG...",
        }

    def dj_query_route(self):
        fetch_metadata = self.fetch_metadata
        attach_relpath = (fetch_metadata["query"] & self.restriction).fetch1(
            *fetch_metadata["fetch_args"]
        )
        with open(Path(os.getcwd(), attach_relpath), "rb") as f:
            image_data = f.read()
        os.unlink(Path(os.getcwd(), attach_relpath))
        return send_file(io.BytesIO(image_data), download_name=attach_relpath)


type_map = {
    "basicquery": BasicQuery,
    "plot:plotly:stored_json": PlotPlotlyStoredjsonComponent,
    "table": TableComponent,
    "metadata": MetadataComponent,
    "file:image:attach": FileImageAttachComponent,
    "slider": BasicQuery,
    "dropdown-query": BasicQuery,
}



from pathlib import Path
import os
from envyaml import EnvYAML
import pkg_resources
import json
import re


def populate_api():
    header_template = """# Auto-generated rest api
from .server import app, protected_route
from .interface import _DJConnector, dj
from flask import request
from json import loads
from base64 import b64decode
from datetime import datetime
import inspect
import traceback
import os
try:
    from .component_interface_override import type_map
except (ModuleNotFoundError, ImportError):
    from .component_interface import type_map
"""
    route_template = """

@app.route('{route}', methods=['GET'])
@protected_route
def {method_name}(jwt_payload: dict) -> dict:

    if request.method in {{'GET'}}:
        try:
            component_instance = type_map['{component_type}'](name='{component_name}',
                                                              component_config={component},
                                                              static_config={static_config},
                                                              jwt_payload=jwt_payload)
            return component_instance.{method_name_type}()
        except Exception as e:
            return traceback.format_exc(), 500
"""
    route_template_nologin = """

@app.route('{route}', methods=['GET'])
def {method_name}() -> dict:
    if request.method in {{'GET'}}:
        jwt_payload = dict(
            databaseAddress=os.environ["PHARUS_HOST"],
            username=os.environ["PHARUS_USER"],
            password=os.environ["PHARUS_PASSWORD"],
        )
        try:
            component_instance = type_map['{component_type}'](name='{component_name}',
                                                              component_config={component},
                                                              static_config={static_config},
                                                              jwt_payload=jwt_payload)
            return component_instance.{method_name_type}()
        except Exception as e:
            return traceback.format_exc(), 500
"""

    pharus_root = f"{pkg_resources.get_distribution('pharus').module_path}/pharus"
    api_path = f"{pharus_root}/dynamic_api.py"
    spec_path = os.environ.get("PHARUS_SPEC_PATH")
    values_yaml = EnvYAML(Path(spec_path))
    with open(Path(api_path), "w") as f:
        f.write(header_template)
        active_route_template = (
            route_template if values_yaml["SciViz"]["auth"] else route_template_nologin
        )
        if (
            "component_interface" in values_yaml["SciViz"]
            and "override" in values_yaml["SciViz"]["component_interface"]
        ):
            with open(
                Path(pharus_root, "component_interface_override.py"), "w"
            ) as component_interface_override:
                component_interface_override.write(
                    values_yaml["SciViz"]["component_interface"]["override"]
                )

        try:
            from .component_interface_override import type_map
        except (ModuleNotFoundError, ImportError):
            from .component_interface import type_map

        static_config = (
            json.dumps(values_yaml["SciViz"]["component_interface"]["static_variables"])
            if (
                "component_interface" in values_yaml["SciViz"]
                and "static_variables" in values_yaml["SciViz"]["component_interface"]
            )
            else None
        )
        pages = values_yaml["SciViz"]["pages"]
        # Crawl through the yaml file for the routes in the components
        for page in pages.values():
            for grid in page["grids"].values():
                if grid["type"] == "dynamic":
                    f.write(
                        (active_route_template).format(
                            route=grid["route"],
                            method_name=grid["route"].replace("/", ""),
                            component_type="basicquery",
                            component_name="dynamicgrid",
                            component=json.dumps(grid),
                            static_config=static_config,
                            method_name_type="dj_query_route",
                        )
                    )

                for comp_name, comp in (
                    grid["component_templates"]
                    if "component_templates" in grid
                    else grid["components"]
                ).items():
                    if re.match(
                        r"^(table|metadata|plot|file|slider|dropdown-query).*$",
                        comp["type"],
                    ):
                        f.write(
                            (active_route_template).format(
                                route=comp["route"],
                                method_name=comp["route"].replace("/", ""),
                                component_type=comp["type"],
                                component_name=comp_name,
                                component=json.dumps(comp),
                                static_config=static_config,
                                method_name_type="dj_query_route",
                            )
                        )
                        if type_map[comp["type"]].attributes_route_format:
                            attributes_route = type_map[
                                comp["type"]
                            ].attributes_route_format.format(route=comp["route"])
                            f.write(
                                (active_route_template).format(
                                    route=attributes_route,
                                    method_name=attributes_route.replace("/", ""),
                                    component_type=comp["type"],
                                    component_name=comp_name,
                                    component=json.dumps(comp),
                                    static_config=static_config,
                                    method_name_type="attributes_route",
                                )
                            )



#! /usr/bin/env python
from setuptools import setup, find_packages
from os import path, listdir
import re

pkg_name = "pharus"
here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.rst"), "r") as f:
    long_description = f.read()

with open(path.join(here, pkg_name, "version.py")) as f:
    exec(f.read())

with open(path.join(here, "requirements.txt")) as f:
    requirements = [
        "{pkg} @ {target}#egg={pkg}".format(
            pkg=re.search(r"/([A-Za-z0-9\-]+)\.git", r).group(1), target=r
        )
        if "+" in r
        else r
        for r in f.read().splitlines()
        if "#" not in r
    ]

setup(
    name=pkg_name,
    version=__version__,
    author="DataJoint Neuro",
    author_email="support@vathes.com",
    description="A generic REST API server backend for DataJoint pipelines.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/datajoint/pharus",
    packages=find_packages(exclude=["test*", "docs"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    entry_points={
        "console_scripts": [f"{pkg_name}={pkg_name}.server:run"],
    },
)



"""Exposed REST API."""
from os import environ
from .interface import _DJConnector, dj
from . import __version__ as version
from typing import Callable
from functools import wraps
from typing import Union

# Crypto libaries
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend as crypto_default_backend

from flask import Flask, request
import jwt
from json import loads
from base64 import b64decode
from datajoint.errors import IntegrityError
from datajoint.table import foreign_key_error_regexp
from datajoint.utils import to_camel_case

app = Flask(__name__)
# Check if PRIVATE_KEY and PUBIC_KEY is set, if not generate them.
# NOTE: For web deployment, please set the these enviorment variable to be the same between
# the instance
if (
    environ.get("PHARUS_PRIVATE_KEY") is None
    or environ.get("PHARUS_PUBLIC_KEY") is None
):
    key = rsa.generate_private_key(
        backend=crypto_default_backend(), public_exponent=65537, key_size=2048
    )
    environ["PHARUS_PRIVATE_KEY"] = key.private_bytes(
        crypto_serialization.Encoding.PEM,
        crypto_serialization.PrivateFormat.PKCS8,
        crypto_serialization.NoEncryption(),
    ).decode()
    environ["PHARUS_PUBLIC_KEY"] = (
        key.public_key()
        .public_bytes(
            crypto_serialization.Encoding.OpenSSH,
            crypto_serialization.PublicFormat.OpenSSH,
        )
        .decode()
    )


def protected_route(function: Callable) -> Callable:
    """
    Protected route function decorator which authenticates requests.

    :param function: Function to decorate, typically routes
    :type function: :class:`~typing.Callable`
    :return: Function output if JWT authetication is successful, otherwise return error
        message
    :rtype: :class:`~typing.Callable`
    """

    @wraps(function)
    def wrapper(**kwargs):
        try:
            jwt_payload = jwt.decode(
                request.headers.get("Authorization").split()[1],
                environ["PHARUS_PUBLIC_KEY"],
                algorithms="RS256",
            )
            return function(jwt_payload, **kwargs)
        except Exception as e:
            return str(e), 401

    wrapper.__name__ = function.__name__
    return wrapper


@app.route(f"{environ.get('PHARUS_PREFIX', '')}/version", methods=["GET"])
def api_version() -> str:
    """
    Handler for ``/version`` route.

    :return: API version
    :rtype: str

    .. http:get:: /version

        Route to check server health returning the API version.

        **Example request**:

        .. sourcecode:: http

            GET /version HTTP/1.1
            Host: fakeservices.datajoint.io

        **Example response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Vary: Accept
            Content-Type: application/json

            {
                "version": "0.1.0"
            }

        :statuscode 200: No error.
    """
    if request.method in {"GET", "HEAD"}:
        return dict(version=version)


@app.route(f"{environ.get('PHARUS_PREFIX', '')}/login", methods=["POST"])
def login() -> dict:
    """
    **WARNING**: Currently, this implementation exposes user database credentials as plain
    text in POST body once and stores it within a bearer token as Base64 encoded for
    subsequent requests. That is how the server is able to submit queries on user's behalf.
    Due to this, it is required that remote hosts expose the server only under HTTPS to ensure
    end-to-end encryption. Sending passwords in plain text over HTTPS in POST request body is
    common and utilized by companies such as GitHub (2021) and Chase Bank (2021). On server
    side, there is no caching, logging, or storage of received passwords or tokens and thus
    available only briefly in memory. This means the primary vulnerable point is client side.
    Users should be responsible with their passwords and bearer tokens treating them as
    one-in-the-same. Be aware that if your client system happens to be compromised, a bad
    actor could monitor your outgoing network requests and capture/log your credentials.
    However, in such a terrible scenario, a bad actor would not only collect credentials for
    your DataJoint database but also other sites such as github.com, chase.com, etc. Please be
    responsible and vigilant with credentials and tokens on client side systems. Improvements
    to the above strategy is currently being tracked in
    https://github.com/datajoint/pharus/issues/82.

    Handler for ``/login`` route.

    :return: Function output is an encoded JWT if successful, otherwise return error message
    :rtype: dict

    .. http:post:: /login

        Route to generate an authentication token.

        **Example request**:

        .. sourcecode:: http

            POST /login HTTP/1.1
            Host: fakeservices.datajoint.io
            Accept: application/json

            {
                "databaseAddress": "tutorial-db.datajoint.io",
                "username": "user1",
                "password": "password1"
            }

        **Example successful response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Vary: Accept
            Content-Type: application/json

            {
                "jwt": "<token>"
            }


        **Example unexpected response**:

        .. sourcecode:: http

            HTTP/1.1 500 Internal Server Error
            Vary: Accept
            Content-Type: text/plain

            400 Bad Request: The browser (or proxy) sent a request that this server could not
                understand.

        :resheader Content-Type: text/plain, application/json
        :statuscode 200: No error.
        :statuscode 500: Unexpected error encountered. Returns the error message as a string.
    """
    if request.method == "POST":
        # Check if request.json has the correct fields
        if not request.json.keys() >= {"databaseAddress", "username", "password"}:
            return dict(error="Invalid json body")

        # Try to login in with the database connection info, if true then create jwt key
        try:
            _DJConnector._attempt_login(
                request.json["databaseAddress"],
                request.json["username"],
                request.json["password"],
            )
            # Generate JWT key and send it back
            encoded_jwt = jwt.encode(
                request.json, environ["PHARUS_PRIVATE_KEY"], algorithm="RS256"
            )
            return dict(jwt=encoded_jwt)
        except Exception as e:
            return str(e), 500


@app.route(f"{environ.get('PHARUS_PREFIX', '')}/schema", methods=["GET"])
@protected_route
def schema(jwt_payload: dict) -> dict:
    """
    Handler for ``/schema`` route.

    :param jwt_payload: Dictionary containing databaseAddress, username, and password strings.
    :type jwt_payload: dict
    :return: If successful then sends back a list of schemas names otherwise returns error.
    :rtype: dict

    .. http:get:: /schema

        Route to get list of schemas.

        **Example request**:

        .. sourcecode:: http

            GET /schema HTTP/1.1
            Host: fakeservices.datajoint.io
            Authorization: Bearer <token>

        **Example successful response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Vary: Accept
            Content-Type: application/json

            {
                "schemaNames": [
                    "alpha_company"
                ]
            }

        **Example unexpected response**:

        .. sourcecode:: http

            HTTP/1.1 500 Internal Server Error
            Vary: Accept
            Content-Type: text/plain

            400 Bad Request: The browser (or proxy) sent a request that this server could not
                understand.

        :reqheader Authorization: Bearer <OAuth2_token>
        :resheader Content-Type: text/plain, application/json
        :statuscode 200: No error.
        :statuscode 500: Unexpected error encountered. Returns the error message as a string.
    """
    if request.method in {"GET", "HEAD"}:
        # Get all the schemas
        try:
            schemas_name = _DJConnector._list_schemas(jwt_payload)
            return dict(schemaNames=schemas_name)
        except Exception as e:
            return str(e), 500


@app.route(
    f"{environ.get('PHARUS_PREFIX', '')}/schema/<schema_name>/table", methods=["GET"]
)
@protected_route
def table(jwt_payload: dict, schema_name: str) -> dict:
    """
    Handler for ``/schema/{schema_name}/table`` route.

    :param jwt_payload: Dictionary containing databaseAddress, username, and password strings.
    :type jwt_payload: dict
    :param schema_name: Schema name.
    :type schema_name: str
    :return: If successful then sends back a list of table names otherwise returns error.
    :rtype: dict

    .. http:get:: /schema/{schema_name}/table

        Route to get tables within a schema.

        **Example request**:

        .. sourcecode:: http

            GET /schema/alpha_company/table HTTP/1.1
            Host: fakeservices.datajoint.io
            Authorization: Bearer <token>

        **Example successful response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Vary: Accept
            Content-Type: application/json

            {
                "tableTypes": {
                    "computed": [],
                    "imported": [],
                    "lookup": [
                        "Employee"
                    ],
                    "manual": [
                        "Computer"
                    ],
                    "part": []
                }
            }

        **Example unexpected response**:

        .. sourcecode:: http

            HTTP/1.1 500 Internal Server Error
            Vary: Accept
            Content-Type: text/plain

            400 Bad Request: The browser (or proxy) sent a request that this server could not
                understand.

        :query schema_name: Schema name.
        :reqheader Authorization: Bearer <OAuth2_token>
        :resheader Content-Type: text/plain, application/json
        :statuscode 200: No error.
        :statuscode 500: Unexpected error encountered. Returns the error message as a string.
    """
    if request.method in {"GET", "HEAD"}:
        try:
            tables_dict_list = _DJConnector._list_tables(jwt_payload, schema_name)
            return dict(tableTypes=tables_dict_list)
        except Exception as e:
            return str(e), 500


@app.route(
    f"{environ.get('PHARUS_PREFIX', '')}/schema/<schema_name>/table/<table_name>/record",
    methods=["GET", "POST", "PATCH", "DELETE"],
)
@protected_route
def record(
    jwt_payload: dict, schema_name: str, table_name: str
) -> Union[dict, str, tuple]:
    (
        """
        Handler for ``/schema/{schema_name}/table/{table_name}/record`` route.

        :param jwt_payload: Dictionary containing databaseAddress, username, and password
            strings.
        :type jwt_payload: dict
        :param schema_name: Schema name.
        :type schema_name: str
        :param table_name: Table name.
        :type table_name: str
        :return: If successful performs desired operation based on HTTP method, otherwise
            returns error.
        :rtype: :class:`~typing.Union[dict, str, tuple]`

        .. http:get:: /schema/{schema_name}/table/{table_name}/record

            Route to fetch records.

            **Example request**:

            .. sourcecode:: http

                GET /schema/alpha_company/table/Computer/record?limit=1&page=2&"""
        "order=computer_id%20DESC&restriction=W3siYXR0cmlidXRlTmFtZSI6ICJjb21wdXRlcl9tZW1vcnk"
        "iLCAib3BlcmF0aW9uIjogIj49IiwgInZhbHVlIjogMTZ9XQo="
        """ HTTP/1.1
                Host: fakeservices.datajoint.io
                Authorization: Bearer <token>

            **Example successful response**:

            .. sourcecode:: http

                HTTP/1.1 200 OK
                Vary: Accept
                Content-Type: application/json

                {
                    "recordHeader": [
                        "computer_id",
                        "computer_serial",
                        "computer_brand",
                        "computer_built",
                        "computer_processor",
                        "computer_memory",
                        "computer_weight",
                        "computer_cost",
                        "computer_preowned",
                        "computer_purchased",
                        "computer_updates",
                        "computer_accessories"
                    ],
                    "records": [
                        [
                            "4e41491a-86d5-4af7-a013-89bde75528bd",
                            "DJS1JA17G",
                            "Dell",
                            1590364800,
                            2.2,
                            16,
                            4.4,
                            "700.99",
                            0,
                            1603181061,
                            null,
                            "=BLOB="
                        ]
                    ],
                    "totalCount": 2
                }

            **Example unexpected response**:

            .. sourcecode:: http

                HTTP/1.1 500 Internal Server Error
                Vary: Accept
                Content-Type: text/plain

                400 Bad Request: The browser (or proxy) sent a request that this server could
                    not understand.

            :query schema_name: Schema name.
            :query table_name: Table name.
            :query limit: Limit of how many records per page. Defaults to ``1000``.
            :query page: Page requested. Defaults to ``1``.
            :query order: Sort order. Defaults to ``KEY ASC``.
            :query restriction: Base64-encoded ``AND`` sequence of restrictions. For example,
                you could restrict as ``[{"attributeName": "computer_memory", "operation": ``-
                ``">=", "value": 16}]`` with this param set as
                ``W3siYXR0cmlidXRlTmFtZSI6ICJjb21wdXRlcl9tZW1vcnkiLCAib3Bl``-
                ``cmF0aW9uIjogIj49IiwgInZhbHVlIjogMTZ9XQo=``. Defaults to no restriction.
            :reqheader Authorization: Bearer <OAuth2_token>
            :resheader Content-Type: text/plain, application/json
            :statuscode 200: No error.
            :statuscode 500: Unexpected error encountered. Returns the error message as a
                string.

        .. http:post:: /schema/{schema_name}/table/{table_name}/record

            Route to insert a record. Omitted attributes utilize the default if set.

            **Example request**:

            .. sourcecode:: http

                POST /schema/alpha_company/table/Computer/record HTTP/1.1
                Host: fakeservices.datajoint.io
                Accept: application/json
                Authorization: Bearer <token>

                {
                    "records": [
                        {
                            "computer_id": "ffffffff-86d5-4af7-a013-89bde75528bd",
                            "computer_serial": "ZYXWVEISJ",
                            "computer_brand": "HP",
                            "computer_built": "2021-01-01",
                            "computer_processor": 2.7,
                            "computer_memory": 32,
                            "computer_weight": 3.7,
                            "computer_cost": 599.99,
                            "computer_preowned": 0,
                            "computer_purchased": "2021-02-01 13:00:00",
                            "computer_updates": 0
                        }
                    ]
                }

            **Example successful response**:

            .. sourcecode:: http

                HTTP/1.1 200 OK
                Vary: Accept
                Content-Type: text/plain

                Insert Successful

            **Example unexpected response**:

            .. sourcecode:: http

                HTTP/1.1 500 Internal Server Error
                Vary: Accept
                Content-Type: text/plain

                400 Bad Request: The browser (or proxy) sent a request that this server could
                    not understand.

            :reqheader Authorization: Bearer <OAuth2_token>
            :resheader Content-Type: text/plain
            :statuscode 200: No error.
            :statuscode 500: Unexpected error encountered. Returns the error message as a
                string.

        .. http:patch:: /schema/{schema_name}/table/{table_name}/record

            Route to update a record. Omitted attributes utilize the default if set.

            **Example request**:

            .. sourcecode:: http

                PATCH /schema/alpha_company/table/Computer/record HTTP/1.1
                Host: fakeservices.datajoint.io
                Accept: application/json
                Authorization: Bearer <token>

                {
                    "records": [
                        {
                            "computer_id": "ffffffff-86d5-4af7-a013-89bde75528bd",
                            "computer_serial": "ZYXWVEISJ",
                            "computer_brand": "HP",
                            "computer_built": "2021-01-01",
                            "computer_processor": 2.7,
                            "computer_memory": 32,
                            "computer_weight": 3.7,
                            "computer_cost": 601.01,
                            "computer_preowned": 0,
                            "computer_purchased": "2021-02-01 13:00:00",
                            "computer_updates": 0
                        }
                    ]
                }

            **Example successful response**:

            .. sourcecode:: http

                HTTP/1.1 200 OK
                Vary: Accept
                Content-Type: text/plain

                Update Successful

            **Example unexpected response**:

            .. sourcecode:: http

                HTTP/1.1 500 Internal Server Error
                Vary: Accept
                Content-Type: text/plain

                400 Bad Request: The browser (or proxy) sent a request that this server could
                    not understand.

            :reqheader Authorization: Bearer <OAuth2_token>
            :resheader Content-Type: text/plain
            :statuscode 200: No error.
            :statuscode 500: Unexpected error encountered. Returns the error message as a
                string.

        .. http:delete:: /schema/{schema_name}/table/{table_name}/record

            Route to delete a specific record.

            **Example request**:

            .. sourcecode:: http

                DELETE /schema/alpha_company/table/Computer/record?cascade=false&"""
        "restriction=W3siYXR0cmlidXRlTmFtZSI6ICJjb21wdXRlcl9tZW1vcnkiLCAib3BlcmF0aW9uIjogIj49"
        "IiwgInZhbHVlIjogMTZ9XQo="
        """ HTTP/1.1
                Host: fakeservices.datajoint.io
                Authorization: Bearer <token>

            **Example successful response**:

            .. sourcecode:: http

                HTTP/1.1 200 OK
                Vary: Accept
                Content-Type: text/plain

                Delete Successful

            **Example conflict response**:

            .. sourcecode:: http

                HTTP/1.1 409 Conflict
                Vary: Accept
                Content-Type: application/json

                {
                    "error": "IntegrityError",
                    "error_msg": "Cannot delete or update a parent row: a foreign key
                        constraint fails (`alpha_company`.`#employee`, CONSTRAINT
                        `#employee_ibfk_1` FOREIGN KEY (`computer_id`) REFERENCES `computer`
                        (`computer_id`) ON DELETE RESTRICT ON UPDATE CASCADE",
                    "child_schema": "alpha_company",
                    "child_table": "Employee"
                }

            **Example unexpected response**:

            .. sourcecode:: http

                HTTP/1.1 500 Internal Server Error
                Vary: Accept
                Content-Type: text/plain

                400 Bad Request: The browser (or proxy) sent a request that this server could
                    not understand.

            :query cascade: Enable cascading delete. Accepts ``true`` or ``false``.
                Defaults to ``false``.
            :query restriction: Base64-encoded ``AND`` sequence of restrictions. For example,
                you could restrict as ``[{"attributeName": "computer_memory", "operation": ``-
                ``">=", "value": 16}]`` with this param set as
                ``W3siYXR0cmlidXRlTmFtZSI6ICJjb21wdXRlcl9tZW1vcnkiLCAib3Bl``-
                ``cmF0aW9uIjogIj49IiwgInZhbHVlIjogMTZ9XQo=``. Defaults to no restriction.
            :reqheader Authorization: Bearer <OAuth2_token>
            :resheader Content-Type: text/plain, application/json
            :statuscode 200: No error.
            :statuscode 409: Attempting to delete a record with dependents while ``cascade``
                set to ``false``.
            :statuscode 500: Unexpected error encountered. Returns the error message as a
                string.
        """
    )
    if request.method in {"GET", "HEAD"}:
        try:
            _DJConnector._set_datajoint_config(jwt_payload)

            schema_virtual_module = dj.VirtualModule(schema_name, schema_name)

            # Get table object from name
            dj_table = _DJConnector._get_table_object(schema_virtual_module, table_name)

            record_header, table_tuples, total_count = _DJConnector._fetch_records(
                query=dj_table,
                **{
                    k: (
                        int(v)
                        if k in ("limit", "page")
                        else (
                            v.split(",")
                            if k == "order"
                            else loads(b64decode(v.encode("utf-8")).decode("utf-8"))
                        )
                    )
                    for k, v in request.args.items()
                },
            )
            return dict(
                recordHeader=record_header, records=table_tuples, totalCount=total_count
            )
        except Exception as e:
            return str(e), 500
    elif request.method == "POST":
        try:
            # Attempt to insert
            _DJConnector._insert_tuple(
                jwt_payload, schema_name, table_name, request.json["records"]
            )
            return "Insert Successful"
        except Exception as e:
            return str(e), 500
    elif request.method == "PATCH":
        try:
            # Attempt to insert
            _DJConnector._update_tuple(
                jwt_payload, schema_name, table_name, request.json["records"]
            )
            return "Update Successful"
        except Exception as e:
            return str(e), 500
    elif request.method == "DELETE":
        try:
            # Attempt to delete tuple
            _DJConnector._delete_records(
                jwt_payload,
                schema_name,
                table_name,
                **{
                    k: loads(b64decode(v.encode("utf-8")).decode("utf-8"))
                    for k, v in request.args.items()
                    if k == "restriction"
                },
                **{
                    k: v.lower() == "true"
                    for k, v in request.args.items()
                    if k == "cascade"
                },
            )
            return "Delete Sucessful"
        except IntegrityError as e:
            match = foreign_key_error_regexp.match(e.args[0])
            return (
                dict(
                    error=e.__class__.__name__,
                    errorMessage=str(e),
                    childSchema=match.group("child").split(".")[0][1:-1],
                    childTable=to_camel_case(match.group("child").split(".")[1][1:-1]),
                ),
                409,
            )
        except Exception as e:
            return str(e), 500


@app.route(
    f"{environ.get('PHARUS_PREFIX', '')}/schema/<schema_name>/table/<table_name>/definition",
    methods=["GET"],
)
@protected_route
def definition(jwt_payload: dict, schema_name: str, table_name: str) -> str:
    """
    Handler for ``/schema/{schema_name}/table/{table_name}/definition`` route.

    :param jwt_payload: Dictionary containing databaseAddress, username, and password strings.
    :type jwt_payload: dict
    :param schema_name: Schema name.
    :type schema_name: str
    :param table_name: Table name.
    :type table_name: str
    :return: If successful then sends back definition for table otherwise returns error.
    :rtype: str

    .. http:get:: /schema/{schema_name}/table/{table_name}/definition

        Route to get DataJoint table definition.

        **Example request**:

        .. sourcecode:: http

            GET /schema/alpha_company/table/Computer/definition HTTP/1.1
            Host: fakeservices.datajoint.io
            Authorization: Bearer <token>

        **Example successful response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Vary: Accept
            Content-Type: text/plain

            # Computers that belong to the company
            computer_id          : uuid                      # unique id
            ---
            computer_serial="ABC101" : varchar(9)            # manufacturer serial number
            computer_brand       : enum('HP','Dell')         # manufacturer brand
            computer_built       : date                      # manufactured date
            computer_processor   : double                    # processing power in GHz
            computer_memory      : int                       # RAM in GB
            computer_weight      : float                     # weight in lbs
            computer_cost        : decimal(6,2)              # purchased price
            computer_preowned    : tinyint                   # purchased as new or used
            computer_purchased   : datetime                  # purchased date and time
            computer_updates=null : time                     # scheduled daily update timeslot
            computer_accessories=null : longblob             # included additional accessories

        **Example unexpected response**:

        .. sourcecode:: http

            HTTP/1.1 500 Internal Server Error
            Vary: Accept
            Content-Type: text/plain

            400 Bad Request: The browser (or proxy) sent a request that this server could not
                understand.

        :reqheader Authorization: Bearer <OAuth2_token>
        :resheader Content-Type: text/plain
        :statuscode 200: No error.
        :statuscode 500: Unexpected error encountered. Returns the error message as a string.
    """
    if request.method in {"GET", "HEAD"}:
        try:
            table_definition = _DJConnector._get_table_definition(
                jwt_payload, schema_name, table_name
            )
            return table_definition
        except Exception as e:
            return str(e), 500


@app.route(
    f"{environ.get('PHARUS_PREFIX', '')}/schema/<schema_name>/table/<table_name>/attribute",
    methods=["GET"],
)
@protected_route
def attribute(jwt_payload: dict, schema_name: str, table_name: str) -> dict:
    """
    Handler for ``/schema/{schema_name}/table/{table_name}/attribute`` route.

    :param jwt_payload: Dictionary containing databaseAddress, username, and password strings.
    :type jwt_payload: dict
    :param schema_name: Schema name.
    :type schema_name: str
    :param table_name: Table name.
    :type table_name: str
    :return: If successful then sends back dict of table attributes otherwise returns error.
    :rtype: dict

    .. http:GET:: /schema/{schema_name}/table/{table_name}/attribute

        Route to get metadata on table attributes.

        **Example request**:

        .. sourcecode:: http

            GET /schema/alpha_company/table/Computer/attribute HTTP/1.1
            Host: fakeservices.datajoint.io
            Authorization: Bearer <token>

        **Example successful response**:

        .. sourcecode:: http

            HTTP/1.1 200 OK
            Vary: Accept
            Content-Type: application/json

            {
                "attributeHeader": [
                    "name",
                    "type",
                    "nullable",
                    "default",
                    "autoincrement"
                ],
                "attributes": {
                    "primary": [
                        [
                            "computer_id",
                            "uuid",
                            false,
                            null,
                            false
                        ]
                    ],
                    "secondary": [
                        [
                            "computer_serial",
                            "varchar(9)",
                            false,
                            "\"ABC101\"",
                            false
                        ],
                        [
                            "computer_brand",
                            "enum('HP','Dell')",
                            false,
                            null,
                            false
                        ],
                        [
                            "computer_built",
                            "date",
                            false,
                            null,
                            false
                        ],
                        [
                            "computer_processor",
                            "double",
                            false,
                            null,
                            false
                        ],
                        [
                            "computer_memory",
                            "int",
                            false,
                            null,
                            false
                        ],
                        [
                            "computer_weight",
                            "float",
                            false,
                            null,
                            false
                        ],
                        [
                            "computer_cost",
                            "decimal(6,2)",
                            false,
                            null,
                            false
                        ],
                        [
                            "computer_preowned",
                            "tinyint",
                            false,
                            null,
                            false
                        ],
                        [
                            "computer_purchased",
                            "datetime",
                            false,
                            null,
                            false
                        ],
                        [
                            "computer_updates",
                            "time",
                            true,
                            "null",
                            false
                        ],
                        [
                            "computer_accessories",
                            "longblob",
                            true,
                            "null",
                            false
                        ]
                    ]
                }
            }

        **Example unexpected response**:

        .. sourcecode:: http

            HTTP/1.1 500 Internal Server Error
            Vary: Accept
            Content-Type: text/plain

            400 Bad Request: The browser (or proxy) sent a request that this server could not
                understand.

        :reqheader Authorization: Bearer <OAuth2_token>
        :resheader Content-Type: text/plain, application/json
        :statuscode 200: No error.
        :statuscode 500: Unexpected error encountered. Returns the error message as a string.
    """
    if request.method in {"GET", "HEAD"}:
        try:
            _DJConnector._set_datajoint_config(jwt_payload)
            local_values = locals()
            local_values[schema_name] = dj.VirtualModule(schema_name, schema_name)

            # Get table object from name
            dj_table = _DJConnector._get_table_object(
                local_values[schema_name], table_name
            )

            attributes_meta = _DJConnector._get_attributes(dj_table)
            return dict(
                attributeHeaders=attributes_meta["attribute_headers"],
                attributes=attributes_meta["attributes"],
            )
        except Exception as e:
            return str(e), 500


@app.route(
    f"{environ.get('PHARUS_PREFIX', '')}/schema/<schema_name>/table/<table_name>/dependency",
    methods=["GET"],
)
@protected_route
def dependency(jwt_payload: dict, schema_name: str, table_name: str) -> dict:
    (
        """
        Handler for ``/schema/{schema_name}/table/{table_name}/dependency`` route.

        :param jwt_payload: Dictionary containing databaseAddress, username, and password
            strings.
        :type jwt_payload: dict
        :param schema_name: Schema name.
        :type schema_name: str
        :param table_name: Table name.
        :type table_name: str
        :return: If sucessfuly sends back a list of dependencies otherwise returns error.
        :rtype: dict

        .. http:get:: /schema/{schema_name}/table/{table_name}/dependency

            Route to get the metadata in relation to the dependent records associated with """
        """a restricted subset of a table.

            **Example request**:

            .. sourcecode:: http

                GET /schema/alpha_company/table/Computer/dependency?restriction=W3siYXR0cml"""
        "idXRlTmFtZSI6ICJjb21wdXRlcl9tZW1vcnkiLCAib3BlcmF0aW9uIjogIj49IiwgInZhbHVlIjogMTZ9XQo"
        "="
        """ HTTP/1.1
                Host: fakeservices.datajoint.io
                Authorization: Bearer <token>

            **Example successful response**:

            .. sourcecode:: http

                HTTP/1.1 200 OK
                Vary: Accept
                Content-Type: application/json

                {
                    "dependencies": [
                        {
                            "accessible": true,
                            "count": 2,
                            "schema": "alpha_company",
                            "table": "computer"
                        },
                        {
                            "accessible": true,
                            "count": 2,
                            "schema": "alpha_company",
                            "table": "#employee"
                        }
                    ]
                }

            **Example unexpected response**:

            .. sourcecode:: http

                HTTP/1.1 500 Internal Server Error
                Vary: Accept
                Content-Type: text/plain

                400 Bad Request: The browser (or proxy) sent a request that this server could
                    not understand.

            :query schema_name: Schema name.
            :query table_name: Table name.
            :query restriction: Base64-encoded ``AND`` sequence of restrictions. For example,
                you could restrict as ``[{"attributeName": "computer_memory", "operation": ``-
                ``">=", "value": 16}]`` with this param set as
                ``W3siYXR0cmlidXRlTmFtZSI6ICJjb21wdXRlcl9tZW1vcnkiLCAib3Bl``-
                ``cmF0aW9uIjogIj49IiwgInZhbHVlIjogMTZ9XQo=``. Defaults to no restriction.
            :reqheader Authorization: Bearer <OAuth2_token>
            :resheader Content-Type: text/plain, application/json
            :statuscode 200: No error.
            :statuscode 500: Unexpected error encountered. Returns the error message as a
                string.
        """
    )
    if request.method in {"GET", "HEAD"}:
        # Get dependencies
        try:
            dependencies = _DJConnector._record_dependency(
                jwt_payload,
                schema_name,
                table_name,
                loads(
                    b64decode(request.args.get("restriction").encode("utf-8")).decode(
                        "utf-8"
                    )
                ),
            )
            return dict(dependencies=dependencies)
        except Exception as e:
            return str(e), 500


def run():
    """
    Starts API server.
    """
    app.run(host="0.0.0.0", port=environ.get("PHARUS_PORT", 5000))


if __name__ == "__main__":
    run()
